require 'yaml'

# Exception class
class YamlEditorError < StandardError; end

# Editor for YAML documents
# rubocop:disable Metrics/ClassLength
class YamlEditor
  # Legal indent chars
  INDENTS = [' ', "\t"].freeze
  COMMENTS = ['#'].freeze
  QUOTES = ['\'', '"'].freeze
  MIX_ERR = [YamlEditorError, 'Mixed indents'].freeze
  VAL_ERR = [YamlEditorError, 'Value has mismatched quotes'].freeze
  SET_ERR = [YamlEditorError,
             'We currently only support adding to hashes'].freeze
  FLD_ERR = [YamlEditorError, 'Values with spaces must be quoted'].freeze

  def initialize(yaml)
    @doc = yaml.lines.map(&:chomp)
  end

  def to_yaml
    @doc.join("\n") + "\n"
  end

  alias to_s to_yaml

  # High level APIs udpate

  def update(*fields, val:)
    raise(*FLD_ERR) if bad_quoting?(val)
    spos, _, kind, v = bracket(*fields)
    if kind == :value
      tag, _, comment = v
      @doc[spos] = "#{tag}#{val}#{comment}"
    else
      append(fields, val)
    end
  end

  # Mid level search API (query)

  # Iterate through a set of fields to obtain the block
  # rubocop:disable all Hookup
  def bracket(*fields)
    spos = 0
    epos = @doc.size - 1
    value = nil
    # Top level document type
    kind = @doc[1] =~ /\s+-/ ? :array : :hash
    fields.each do |field|
      if field.is_a?(Integer) && kind == :array
        spos, epos = array(spos, epos)[field]
      else
        spos, epos, kind, value = inner_bracket(field, spos, epos)
      end
      break unless spos
    end
    [spos, epos, kind, value]
  end
  # rubocop:enable all

  # Low level public APIs

  # Calculate a block start and finish based upon a field
  # rubocop:disable Metrics/MethodLength  ok
  # rubocop:disable Metrics/AbcSize       ok
  def inner_bracket(field, spos, epos)
    ind = @doc[spos..epos].index { |v| v =~ /\A\s*-*\s*#{field}\:/ }
    return unless ind
    ind += spos
    start_line = @doc[ind]
    schar, snum = indented(start_line)
    val = value(start_line)
    return [ind, ind, :value, val] if val # Value returned if value
    ind += 1
    kind = array?(@doc[ind]) ? :array : :hash
    # Loop until we encounter same indent level or end
    @doc[ind..epos].each_with_index do |line, eind|
      nchar, nnum = indented(line)
      raise(*MIX_ERR) if nchar != schar
      return [ind, ind + eind - 1, kind] if nnum <= snum
    end
    [ind, epos, kind]
  end
  # rubocop:enable all

  # Produce a set of bracketed array lines betweem a line range
  # requires that we start with a known array position containing
  # the hyphen
  # rubocop:disable Metrics/AbcSize ok
  def array(spos, epos)
    # Get initial indent level
    ilvl = non_ws_index(@doc[spos])
    # Produce start and end pairings
    res = @doc[spos..epos].each_with_index.map do |line, ind|
      nlvl = non_ws_index(line)
      nlvl == ilvl ? spos + ind : nil
    end
    # Strip non boundaries, append a final marker and collect range groups
    res.compact.push(epos + 1).each_cons(2).map do |s, f|
      # Skip empty boundaries
      s += 1 until hash_key?(@doc[s])
      s >= f ? nil : [s, f.pred]
    end
       .compact
  end
  # rubocop:enable all

  # Analyse the line for a value
  def value(line)
    res = line.match(/\A(?<tag>.+\:\s)(?<value>.+)\Z/)
    return unless res
    tag = res[:tag]
    val = res[:value]
    val, rest = value_split(val)
    return if val.empty?
    # Return the split line that could be reassembled
    [tag, val, rest]
  end

  # RAW access
  # :nocov:
  def lines(range)
    @doc[range]
  end
  # :nocov:

  private

  # Split the value half to capture comments etc.
  def value_split(val)
    # Comment only ?
    return ['', val] if val.start_with?(*COMMENTS)
    # No quotes ?
    return split_noquote(val) unless val.start_with?(*QUOTES)
    # Quotes
    split_quoted(val)
  end

  # Simple unquoted values
  def split_noquote(val)
    pos = val.index(/\s/)
    pos ? [val[0..pos.pred], val[pos..-1]] : [val, '']
  end

  # Quoted values
  def split_quoted(val)
    prev = quote = val[0]
    pos = 1
    until val[pos] == quote && prev != '\\'
      prev = val[pos]
      pos += 1
      raise(*VAL_ERR) if pos > val.length
    end
    [val[0..pos], val[pos.succ..-1]]
  end

  # Type and count of indent char
  def indented(line)
    spaces = line.match(/\A\s*-*\s*/).to_s
    ichars = INDENTS.map { |c| spaces.count(c) }
    raise(*MIX_ERR) if ichars.count(&:nonzero?) > 1
    # On zero indents we default to space and zero
    iichar = ichars.index(&:nonzero?) || 0
    # Add any hyphens for arrays detected inline
    [INDENTS[iichar], ichars[iichar] + spaces.count('-')]
  end

  # Used by set in cases where the field does not exist
  # rubocop:disable Metrics/AbcSize ok
  def append(fields, val)
    parent = fields.dup
    node = parent.pop
    spos, epos, = bracket(*parent)
    # Must inspect inner element for type
    kind = value(@doc[spos]) ? :hash : :array
    raise(*SET_ERR) unless kind == :hash
    ichar, inum = indented(@doc[epos])
    pad = ichar * inum
    @doc.insert(epos.succ, "#{pad}#{node}: #{val}")
  end
  # rubocop:enable all

  # Are quotes missing but needed?
  def bad_quoting?(val)
    val.include?(' ') && !QUOTES.include?(val[0]) || false
  end

  # Simple position of first non whitespace
  def non_ws_index(line)
    line.index(/[^\s]/)
  end

  # Simple Hash key presence
  def hash_key?(line)
    line =~ /\A\s*-*\s*.+:/
  end

  # Array marker presence
  def array?(line)
    line =~ /\A\s*-/
  end
end
